#ifndef __HTTPCONNECTION_H_
#define __HTTPCONNECTION_H_

#include <windows.h>
#include <tchar.h>
#include <WinInet.h>
#include <vector>
#include <string>
#include <map>

#include "Base64.h"

using namespace std;
typedef basic_string<TCHAR> tstring;

/**
 * HTTPڑNX
 */
class HttpConnection
{

public:
	// G[
	enum Error {
		ERROR_NONE,  // G[Ȃ
		ERROR_INET,  // lbg[NCũG[
		ERROR_CANCEL // LZꂽ
	};

	/**
	 * NGXgpR[obN
	 * @param context ReLXg
	 * @param buffer ݐf[^obt@
	 * @param size ݐf[^obt@̃TCYBۂɏ񂾃TCYi[ĕԂ
	 * @return fꍇ true Ԃ
	 */
	typedef bool (*RequestCallback)(void *context, void *buffer, DWORD &size);


	/**
	 * gCpR[obN
	 * @param context ReLXg
	 * @return fꍇ true Ԃ
	 */
	typedef void (*RetryCallback)(void *context);
	
	/**
	 * X|XpR[obN
	 * @param context ReLXg
	 * @param buffer ǂݍ݌f[^obt@BŌ NULL
	 * @param size f[^obt@̃TCYBŌ0
	 * @return fꍇ true Ԃ
	 */
	typedef bool (*ResponseCallback)(void *context, const void *buffer, DWORD size);
	
	/**
	 * RXgN^
	 * @param agentName G[WFg
	 * @param checkCert F؊mF邩ǂ
	 */
	HttpConnection(tstring agentName, bool checkCert=false) : agentName(agentName), checkCert(checkCert), contentLength(0), secure(false){
		::InitializeCriticalSection(&cs);
		hInet = NULL;
		hConn = NULL;
		hReq  = NULL;
	}

	// fXgN^
	~HttpConnection(void) {
		clearParam();
		::DeleteCriticalSection(&cs);
	}

	// Mwb_NA
	void clearHeader() {
		header.clear();
		requestContentLength = 0;
		requestContentType.erase();
		requestEncoding.erase();
	}

	// nhNA
	void closeHandle();

	// Mp[^NA(OςŁÂ͑Mf[^NA)
	void clearParam() {
		closeHandle();
		clearHeader();
	}

	// ----------------------------------------------------------------------------------------
	
	// HTTP wb_ǉ
	void addHeader(const TCHAR *name, const TCHAR *value);

	// F؃wb_Zbg(addHeader ̃[eBeB)
	void addBasicAuthHeader(const tstring &user, const tstring &passwd) {
		tstring sendStr = user + _T(":") + passwd;
		tstring value = _T("Basic") + base64encode(sendStr.c_str(), sendStr.length());
		addHeader(_T("Authorization"), value.c_str());
	}
	
	// ----------------------------------------------------------------------------------------------------
	
	/**
	 * NGXgJn
	 * @param method ANZX\bh
	 * @param url URL
	 * @param user ANZX[U
	 * @param passwd ANZXpX[h
	 * @return  true
	 */
	bool open(const TCHAR *method,
			  const TCHAR *url,
			  const TCHAR *user = NULL,
			  const TCHAR *passwd = NULL);

	/**
	 * NGXgM
	 * @param callback MpR[obN
	 * @param context R[obNpReLXg
	 * @return G[
	 */
	int request(RequestCallback requestCallback=NULL, RetryCallback retryCalblack = NULL, void *context=NULL);


	/**
	 * X|X擾OW
	 */
	void queryInfo();
	
	/**
	 * X|XM
	 * @param callback ۑpR[obN
	 * @param context R[obNpReLXg
	 * @return G[
	 */
	int response(ResponseCallback callback=NULL, void *context=NULL);
	
	// ----------------------------------------------------------------------------------------------------
	
	// G[bZ[W̎擾
	const TCHAR *getErrorMessage() const {
		return errorMessage.c_str();
	}
	
	// Ō̃NGXgĂ邩ǂ
	bool isValid() const {
		return hReq != NULL;
	}

	// HTTPXe[^XR[h擾
	int getStatusCode() const {
		return statusCode;
	}

	// HTTPXe[^XeLXg擾
	const TCHAR *getStatusText() const {
		return statusText.c_str();
	}
	
	// 擾ꂽRec̒
	int getContentLength() const {
		return contentLength;
	}

	// Rec MIME-TYPE
	const TCHAR *getContentType() const {
		return contentType.c_str();
	}

	// Rec̃GR[fBO
	const TCHAR *getEncoding() const {
		return encoding.c_str();
	}

	// Rec̃GR[fBO
	const TCHAR *getRequestEncoding() const {
		return requestEncoding.c_str();
	}
	
	/**
	 * X|X̃wb_擾
	 */
	const TCHAR *getResponseHeader(const TCHAR *name) {
		map<tstring,tstring>::const_iterator it = responseHeaders.find(tstring(name));
		if (it != responseHeaders.end()) {
			return it->second.c_str();
		}
		return NULL;
	}

	// X|Xwb_S擾p:
	void initRH() {
		rhit = responseHeaders.begin();
	}

	// X|Xwb_S擾p:擾
	bool getNextRH(tstring &name, tstring &value) {
		if (rhit != responseHeaders.end()) {
			name  = rhit->first;
			value = rhit->second;
			rhit++;
			return true;
		}
		return false;
	}

	bool getCheckCert() const { return checkCert;	}
	void setCheckCert(bool check)	{ checkCert = check;}

private:
	CRITICAL_SECTION cs;

	// b
	tstring agentName; ///< [UG[WFg
	bool checkCert;	   ///< ؖmF_CAOo
	bool secure;       ///< https ʐMǂ

	HINTERNET hInet; ///< C^[lbgڑ
	HINTERNET hConn; ///< RlNV
	HINTERNET hReq;  ///< HTTPNGXg
	
	// Mpf[^
	vector<tstring> header;	///< HTTP wb_
	DWORD requestContentLength; ///< NGXg Content-Length:
	tstring requestContentType; ///< NGXg Content-Type:
	tstring requestEncoding;    ///< NGXg̃GR[hw

	// Mpf[^
	bool validContentLength;
	DWORD contentLength;     ///< Content-Length:
	tstring contentType;     ///< Content-Type: type
	tstring encoding;        ///< Content-TYpe: ̃GR[fBO

	DWORD statusCode;        ///< HTTP status code
	tstring statusText;      ///< HTTP status text
	map<tstring,tstring> responseHeaders; ///< X|Xwb_
	map<tstring,tstring>::const_iterator rhit; //< X|Xwb_QƗpCe[^

	// G[R[h
	tstring errorMessage; ///< G[bZ[W
};

#endif
